import time
from typing import List

from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
from selenium.webdriver import ChromeOptions

from .utils import source2response, set_browser_options, apply_settings



class WebDriverMixin:

    def sleep(self, delay=1):
        time.sleep(delay)

    def find_element(self, value, by=By.CSS_SELECTOR) -> WebElement:
        return super().find_element(by, value)
    
    def find_elements(self, value, by=By.CSS_SELECTOR) -> List[WebElement]:
        return super().find_elements(by, value)
    
    def clear_browser_cache(self):
        self.command_executor._commands['SEND_COMMAND'] = ('POST', '/session/$sessionId/chromium/send_command')
        self.execute('SEND_COMMAND', dict(cmd='Network.clearBrowserCache', params={}))
    
    @property
    def response(self):
        return source2response(self.current_url, self.page_source)

    def __del__(self):
        self.quit()



def load_chrome_driver_service():
    manager = ChromeDriverManager()
    driver_path=manager.install()
    return Service(driver_path)


class ChromeDriver(WebDriverMixin, webdriver.Chrome):

    def __init__(self, *args, **kwargs):
        service = load_chrome_driver_service()
        options = set_browser_options(ChromeOptions(), **kwargs)
        super().__init__(*args, service=service, options=options)
        apply_settings(self, **kwargs)




def parse_number(value:str):
    if isinstance(value, (int, float)):
        return value

    value = value.strip('.')
    dot_count = value.count('.')

    if dot_count == 0:
        tokens = filter(str.isdigit, value)
        return int(''.join(tokens))
    elif dot_count == 1:
        tokens = filter(lambda c: c == '.' or c.isdigit(), value)
        return float(''.join(tokens))
    else:
        return value


import os, glob



def get_uniquify_path(path):
    n_path = path
    f, ext = os.path.splitext(path)
    n = 1
    while os.path.exists(n_path):
        n_path = f'{f}({n}){ext}'
        n += 1
    return n_path


def mkdirs(path):
    dir = os.path.dirname(path)
    if not os.path.exists(dir):
        os.makedirs(dir)


def fsave(content, path, exists='overwrite', **kwargs):
    '''ifexists: overwrite|skip|distinct
    '''
    if isinstance(content, str):
        mode = 'wt'
    elif isinstance(content, bytes):
        mode = 'wb'
    else:
        raise ValueError(f'content must be bytes or text object. not {type(content)}')

    mkdirs(path)

    if exists == 'distinct':
        path = get_uniquify_path(path)
    elif exists == 'skip':
        if os.path.exists(path):
            return 0
    elif exists == 'overwrite':
        pass
    else:
        raise ValueError(f'exists must be in overwrite, distinct, skip')

    with open(path, mode, **kwargs) as fp:
        fp.write(content)
    return len(content)


def fcount(pathname, recursive=False):
    files = [f for f in glob.glob(pathname, recursive=recursive)]
    count = len(files)
    return count


def fdirname(p):
    return os.path.dirname(p)


def fbasename(p):
    return os.path.basename(p)


def fextname(p):
    return os.path.splitext(p)[1]


def fexists(path):
    return os.path.exists(path)


import math
from random import random
from typing import Callable, Generator


def gen_random_value(low:int, high:int, rounds=2):
    width = high - low
    v = low+random()*width
    return round(v, rounds)


def format_bytes_size(length, ndigits=None):
    k = 1024
    m = k**2
    g = k**3

    if length > k:
        unit = 'K'
        v = length / k
    elif length > m:
        unit = 'M'
        v = length / m
    elif length > g:
        unit = 'G'
        v = length / g
    else:
        unit = 'bytes'
        v = length
    return f'{round(v, ndigits)}{unit}'


def array2matrix(array:list, n:int, orphan=True):
    length = len(array)
    s, e = 0, n
    result = []
    while e < length:
        slc = array[s:e]
        s = e
        e = e + n
        result.append(slc)

    rest = array[s:e]
    is_orphan = rest and len(rest) < n

    if not is_orphan:
        result.append(rest)
    
    if is_orphan:
        if orphan:
            result.append(rest)
    return result


def get_next_rotate(iterable, current):
    rotater = list(iterable) * 2
    for i, value in enumerate(rotater):
        if current == value:
            return rotater[i+1]
    raise ValueError(f'Cannot find {current} in {iterable}')


def get_rotate(iterable):
    if iterable:
        first, *rest = iterable
        return [*rest, first]
    raise ValueError(f'{iterable}must have at least one element')


def npaginator(page_size, total, start=1, **kwargs):
    last = math.ceil(total / page_size)
    if start == 0:
        yield from range(start, last, **kwargs)
    else:
        yield from range(start, last + 1, **kwargs)
        




from pathlib import Path
from collections import abc

from scrapgo.lib import queryjoin, urljoin
from .base import UrlRenderAction,  FileAction, CurlAction



def resolve_link(action, link, response=None):
    if isinstance(action, FileAction):
        if link.startswith('file:///'):
            return link
        else:
            return Path(link).absolute().as_uri()
            
    if isinstance(action, CurlAction):
        return link

    if isinstance(action, UrlRenderAction): 
        host = action.url or response.url
    else:
        host = response.url

    if isinstance(link, abc.Mapping):
        return queryjoin(host, link)
    return urljoin(host, link)



class BaseCrawlerException(Exception):
    '''base crawler exception
    '''


class CannotFindAction(BaseCrawlerException):
    '''cannot find action
    '''


class CannotParseResponse(BaseCrawlerException):
    '''cannot parsable response
    '''


class FollowNameError(BaseCrawlerException):
    '''follow argument error
    '''

class UrlError(BaseCrawlerException):
    '''UrlError
    '''

class UrlOrderError(BaseCrawlerException):
    '''Url Order Error
    '''


class UrlRendererError(BaseCrawlerException):
    '''url renderer retun value type
    '''


import inspect
from itertools import takewhile
from collections import abc

from .regex import strcompile
from .exceptions import ReduceModuleException



def find_function(module, regex):
    if not inspect.ismodule(module):
        raise ValueError(f"{module} is Not python module  .py")
    for name, func in inspect.getmembers(module, inspect.isfunction):
        if g := strcompile(regex).search(name):
            yield g.group, func


def module2dict(module):
    return dict(takewhile(lambda i: i[0] != '__builtins__', inspect.getmembers(module)))


def reduce_module2dict(module):
    if inspect.ismodule(module):
        return module2dict(module)
    elif isinstance(module, abc.Mapping):
        return module
    else:
        raise ReduceModuleException(
            f"{module} is must be module or mapping"
        )


def get_kwargnames(callable):
    sig = inspect.signature(callable)
    return list(sig.parameters)


def select_kwargs(callable, *args, allowed_params:list=None, **kwargs):
    allowed_params = allowed_params or []
    allowed_params += get_kwargnames(callable)
    kwargs = {
        key: value for key, value in kwargs.items()
        if key in allowed_params
    }
    return callable(*args, **kwargs)




import json
from itertools import dropwhile



def find_json(content, many=True):
    # vars,
    OPENING_BRAKET, CLOSING_BRAKET, OPENING_BRACE, CLOSING_BRACE = 'OPENING_BRAKET', 'CLOSING_BRAKET', 'OPENING_BRACE', 'CLOSING_BRACE',
    OPENING, CLOSING = 'OPENING', 'CLOSING',
    BRAKET, BRACE = 'BRAKET', 'BRACE',
    objmap = {
        '[': OPENING_BRAKET, b'[': OPENING_BRAKET,
        ']': CLOSING_BRAKET, b']': CLOSING_BRAKET,
        '{': OPENING_BRACE, b'{': OPENING_BRACE,
        '}': CLOSING_BRACE, b'}': CLOSING_BRACE,
    }

    balancer = {
        BRAKET: 0, BRACE:0
    }

    ## cut off head, tail impure
    content = ''.join(dropwhile(lambda c: objmap.get(c) not in [OPENING_BRAKET, OPENING_BRACE], content))
    content = ''.join(reversed(list(dropwhile(lambda c: objmap.get(c) not in [CLOSING_BRAKET, CLOSING_BRACE], reversed(content)))))


    _records = []
    for i, c in enumerate(content):
        if not(token := objmap.get(c)):
            continue

        option, kind = token.split('_') # ex) OPENING, BRACE

        if option == OPENING:
            balancer[kind] += 1
        else:
            balancer[kind] -= 1
        
        record = dict(
            index=i, kind=kind, option=option, balance=balancer[kind]
        )
        _records.append(record)
    
    if len(_records) < 2:
        return []

    starts = None
    tokens = []
    for r in _records:
        if starts is None:
            starts = r
            continue
        if starts['kind'] == r['kind'] and r['option'] == CLOSING and r['balance'] == 0:
            tok = content[starts['index']: r['index']+1]
            tokens.append(tok)
            starts = None

    success = []
    for tok in tokens:
        try:
            obj = json.loads(tok)
        except json.decoder.JSONDecodeError:
            continue
        success.append(obj)
    if many:
        return success
    else:
        if success:
            return success[0]



from bs4.element import Tag



def prettier(content):
    if not isinstance(content, (str, Tag, bytes)):
        return content

    if isinstance(content, Tag):
        content = content.get_text(strip=True)
        
    trantab = {
        '<br>': '',
        '</br>': '\n',
        '<p>':'',
        '</p>': '\n',
        '\t': '',
        '\u200b': '',
        '\xa0': '',
        
    }

    for tok, repl in trantab.items():
        content = content.replace(tok, repl)
    
    lines = []
    for line in content.split('\n'):
        if line := line.strip():
            lines.append(line)
    return '\n'.join(lines)




def get_nested_soup_depth(soup, *args, max_depth=True, **kwargs):
    def counter(soup, count=0):
        for e in soup(*args, **kwargs):
            yield from counter(e, count+1)
        yield count
    if max_depth:
        return max(counter(soup))
    return min(counter(soup))



def filter_nested_soup_by_depth(soup, *args, depth=0, **kwargs):
    return [
        e for e in soup(*args, **kwargs)
        if get_nested_soup_depth(e, *args, **kwargs) == depth
    ]


def find_table(soup:Tag, *columns:str, tag_name='table'):
    def filter(tag):
        if tag.name != tag_name:
            return False
        for col in columns:
            if col not in tag.text:
                return False
        return True
    depth = get_nested_soup_depth(soup, filter, max_depth=False)
    if tables := filter_nested_soup_by_depth(soup, filter, depth=depth):
        return tables[0]


def load_form_values(soup:Tag, selector, value_field='value', fill_value=None):
    return {
        input['name']: input.get(value_field, fill_value)
        for input in soup.select_one(selector).select('[name]')
    }
        



from requests_file import FileAdapter



class RequestsFileAdapter(FileAdapter):
    def send(self, request, **kwargs):
        resp = super().send(request, **kwargs)
        resp.request = request
        return resp


class FileRequestsMixin:

    def load_session(self):
        session = super().load_session()
        session.mount('file:///', RequestsFileAdapter())
        return session

    


from mimetypes import guess_type

from selenium import webdriver
from requests.models import Response


def set_browser_options(options, browser_options:list=None, browser_experimental_options:dict=None, **kwargs):
    for opt in browser_options or []:
        options.add_argument(opt)
    for opt, value in (browser_experimental_options or {}).items():
        options.add_experimental_option(opt, value)
    return options


def apply_settings(driver:webdriver, browser_settings:dict=None, **kwargs):
    for conf, values in (browser_settings or {}).items():
        getattr(driver, conf)(*values)


def source2response(url, page_source):
    response = Response()
    response.url = url
    response.status_code = 200
    response._content = page_source.encode('utf-8')
    response.headers = {
        'Content-Type': guess_type(url)
    }
    return response

